from typing import Any, Dict, List

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from toolz import concat

from saiph.reduction import DUMMIES_PREFIX_SEP
from saiph.reduction.utils.common import (
    column_multiplication,
    get_dummies_mapping,
    get_explained_variance,
    get_grouped_modality_values,
    row_division,
    row_multiplication,
)


@pytest.fixture
def df() -> pd.DataFrame:
    return pd.DataFrame([[1, 10], [2, 20]])


def test_row_multiplication(df: pd.DataFrame) -> None:
    expected = pd.DataFrame([[1, 10], [4, 40]])

    result = row_multiplication(df, np.array([1, 2]))

    assert_frame_equal(result, expected)


def test_column_multiplication(df: pd.DataFrame) -> None:
    expected = pd.DataFrame([[1, 20], [2, 40]])

    result = column_multiplication(df, np.array([1, 2]))
    assert_frame_equal(result, expected)


def test_row_division(df: pd.DataFrame) -> None:
    expected = pd.DataFrame([[1, 10], [1, 10]], dtype=float)

    result = row_division(df, np.array([1, 2]))

    assert_frame_equal(result, expected)


def test_get_dummies_mapping(
    quali_df: pd.DataFrame, mapping: Dict[str, List[str]]
) -> None:
    dummy_columns = pd.get_dummies(quali_df, prefix_sep=DUMMIES_PREFIX_SEP).columns
    result = get_dummies_mapping(quali_df.columns, dummy_columns)

    assert result == mapping


def test_get_grouped_modality_values(mapping: Dict[str, List[str]]) -> None:
    """Verify that grouping modalities returns the correct groupings."""
    df = pd.DataFrame.from_dict(
        {col: [10] for col in concat(mapping.values())}, orient="index"
    )
    expected = pd.DataFrame.from_dict(
        {col: [20] for col in mapping.keys()}, orient="index"
    )
    grouped_df = get_grouped_modality_values(mapping, df)

    assert_frame_equal(grouped_df.sort_index(), expected.sort_index())


@pytest.mark.parametrize(
    "s,expected_variance,expected_ratio",
    [
        ([10, 10, 10], [50.0, 50.0], [1 / 3, 1 / 3]),  # nf < len(s), we don't get 100%.
        ([0.0, 0.0, 0.0], [0.0, 0.0], [np.nan, np.nan]),
    ],
)
def test_get_explained_variance_returns_correct_variance_and_ratio(
    s: Any, expected_variance: Any, expected_ratio: Any
) -> None:
    variance, ratio = get_explained_variance(np.array(s), nb_individuals=3, nf=2)
    np.testing.assert_array_equal(variance, expected_variance)
    np.testing.assert_array_equal(ratio, expected_ratio)



import sys
from typing import Optional

import pandas as pd


def filter_variable_contribution(
    contributions: pd.DataFrame,
    dim: int = 1,
    *,
    max_var: Optional[int] = 10,
    min_contrib: Optional[float] = 0.1,
) -> pd.Series:
    """Get sorted and filtered variable contributions for a given dimension.

    Parameters:
        contributions: Variable contributions of the model, per dimension.
        dim: Dimension to plot. default: 1
        max_var : Maximum number of variables to plot. default: 10
        min_contrib : Minimum contribution threshold for the variable
                    contributions to be displayed. default: 0.1

    Returns:
        selected: Contributions of the specified dimension, sorted by descending importance
            and filtered by the given values of `max_var` and `min_contrib`.
    """
    nb_variables = max_var or len(contributions)
    min_contrib = min_contrib or sys.float_info.min
    label = contributions.columns[dim - 1]  # Dimensions starts at 1
    selected = (
        contributions[label]
        .loc[lambda v: v > min_contrib]
        .head(nb_variables)
        .sort_values(ascending=False)
    )
    return selected



"""Inverse transform coordinates."""
import ast
from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from saiph.exception import InvalidParameterException
from saiph.models import Model
from saiph.reduction import DUMMIES_PREFIX_SEP
from saiph.reduction.utils.common import get_dummies_mapping


def inverse_transform(
    coord: pd.DataFrame,
    model: Model,
    *,
    use_approximate_inverse: bool = False,
    use_max_modalities: bool = True,
    seed: Optional[int] = None,
) -> pd.DataFrame:
    """Return original format dataframe from coordinates.

    Parameters:
        coord: coord of individuals to reverse transform
        model: model used for projection
        use_approximate_inverse: matrix is not invertible when n_individuals < n_dimensions
            an approximation with bias can be done by setting to ``True``. default: ``False``
        use_max_modalities: for each variable, it assigns to the individual
            the modality with the highest proportion (True)
            or a random modality weighted by their proportion (False). default: True
        seed: seed to fix randomness if use_max_modalities = False. default: None

    Returns:
        inverse: coordinates transformed into original space.
            Retains shape, encoding and structure.
    """
    # Check dimension size regarding N
    n_dimensions = len(model.dummy_categorical) + len(model.original_continuous)
    n_records = len(coord)

    if not use_approximate_inverse and n_records < n_dimensions:
        raise InvalidParameterException(
            f"n_dimensions ({n_dimensions}) is greater than n_records ({n_records})."
        )
    # Get back scaled_values from coord with inverse matrix operation
    # If n_records < n_dimensions, There will be an approximation of the inverse of V.T
    scaled_values = pd.DataFrame(coord @ np.linalg.pinv(model.V.T))
    # get number of continuous variables
    nb_quanti = len(model.original_continuous)

    # separate quanti from quali
    scaled_values_quanti = scaled_values.iloc[:, :nb_quanti]
    scaled_values_quanti.columns = model.original_continuous

    scaled_values_quali = scaled_values.iloc[:, nb_quanti:]
    scaled_values_quali.columns = model.dummy_categorical

    # Descale regarding projection type
    # FAMD
    if model.type == "famd":
        descaled_values_quanti = (scaled_values_quanti * model.std) + model.mean
        descaled_values_quali = (scaled_values_quali * np.sqrt(model.prop)) + model.prop
        del scaled_values_quali
        del scaled_values_quanti
        undummy = undummify(
            descaled_values_quali,
            get_dummies_mapping(model.original_categorical, model.dummy_categorical),
            use_max_modalities=use_max_modalities,
            seed=seed,
        )
        inverse = pd.concat([descaled_values_quanti, undummy], axis=1).round(12)

    # PCA
    elif model.type == "pca":
        descaled_values_quanti = (scaled_values_quanti * model.std) + model.mean
        inverse = descaled_values_quanti.round(12)
        del scaled_values_quali
        del scaled_values_quanti

    # MCA
    else:
        del scaled_values_quali
        del scaled_values_quanti
        # As we are not scaling MCA such as FAMD categorical, the descale is
        # not the same. Doing the same as FAMD is incoherent.
        inverse_data = coord @ (model.D_c @ model.V.T).T
        inverse_coord_quali = inverse_data.set_axis(
            model.dummy_categorical, axis="columns"
        )

        descaled_values_quali = inverse_coord_quali.divide(model.dummies_col_prop)
        inverse = undummify(
            descaled_values_quali,
            get_dummies_mapping(model.original_categorical, model.dummy_categorical),
            use_max_modalities=use_max_modalities,
            seed=seed,
        )
    # Cast columns to same type as input
    for name, dtype in model.original_dtypes.iteritems():
        # Can create a bug if a column is object but contains int and float values,
        # first, we force the value type of the first value of the original df
        if dtype in ["object", "category"]:
            if model.modalities_types[name] == "bool":
                inverse[name] = [ast.literal_eval(ele) for ele in inverse[name]]
            else:
                inverse[name] = inverse[name].astype(model.modalities_types[name])

        inverse[name] = inverse[name].astype(dtype)

    # reorder columns
    return inverse[model.original_dtypes.index]


def undummify(
    dummy_df: pd.DataFrame,
    dummies_mapping: Dict[str, List[str]],
    *,
    use_max_modalities: bool = True,
    seed: Optional[int] = None,
) -> pd.DataFrame:
    """Return undummified dataframe from the dummy dataframe.

    Parameters:
        dummy_df: dummy df of categorical variables
        dummies_mapping: mapping between categorical columns and dummies columns.
        use_max_modalities: True to select the modality with the highest probability.
                            False for a weighted random selection. default: True
        seed: seed to fix randomness if use_max_modalities = False. default: None

    Returns:
        inverse_quali: undummify df of categorical variable
    """
    inverse_quali = pd.DataFrame()
    random_gen = np.random.default_rng(seed)

    def get_suffix(string: str) -> str:
        return string.split(DUMMIES_PREFIX_SEP)[1]

    for original_column, dummy_columns in dummies_mapping.items():
        # Handle a single category with all the possible modalities
        single_category = dummy_df[dummy_columns]
        if use_max_modalities:
            # select modalities with highest probability
            chosen_modalities = single_category.idxmax(axis="columns")
        else:
            chosen_modalities = get_random_weighted_columns(single_category, random_gen)
        inverse_quali[original_column] = list(map(get_suffix, chosen_modalities))

    return inverse_quali


def get_random_weighted_columns(
    df: pd.DataFrame, random_gen: np.random.Generator
) -> pd.Series:
    """Randomly select column labels weighted by proportions.

    Parameters:
        df : dataframe containing proportions
        random_gen: random generator

    Returns:
        column_labels: selected column labels
    """
    # Example for 1 row:  [0.1, 0.3, 0.6] --> [0.1, 0.4, 1.0]
    cum_probability = df.cumsum(axis=1)
    random_probability = random_gen.random((cum_probability.shape[0], 1))
    # [0.342] < [0.1, 0.4, 1.0] --> [False, True, True] --> idx: 1
    column_labels = (random_probability < cum_probability).idxmax(axis=1)

    return column_labels



from .inverse_transform import inverse_transform
from .projection import fit, fit_transform, stats, transform

# Also modify in pyproject.toml
__version__ = "1.4.0"

__all__ = [
    "__version__",
    "fit",
    "fit_transform",
    "inverse_transform",
    "transform",
    "stats",
]
